package protocolP2P;
import java.util.Arrays;
import protocolP2P.Payload;
import protocolP2P.RequestResponseCode;
import exception.TransmissionError;
import exception.ProtocolError;
import exception.InternalError;
import exception.SizeError;
import java.io.UnsupportedEncodingException;

/** Representation of payload for list response.
 * @author	Louis Royer
 * @author	Flavien Haas
 * @author	JS Auge
 * @version	1.0
 */
public class FileList extends Payload {
	private String[] fileList;

	/** Constructor (typically used by the server) with an ArrayList parameter containing
	 * filenames.
	 * @param fileList	a list of files. Must not be empty.
	 * @throws	InternalError
	 */
	public FileList(String[] fileList) throws InternalError {
		super(RequestResponseCode.LIST_RESPONSE);
		/* assert to help debugging */
		assert fileList.length != 0 : "Payload size of FileList must not be empty, use EmptyDirectory from Payload instead";
		if (fileList.length == 0) {
			throw new InternalError();
		}
		this.fileList = fileList;
	}

	/** Constructor (typically used by client) with a byte[] parameter containing the datagram received.
	 * @param datagram	the full datagram received
	 * @throws	SizeError
	 * @throws	InternalError
	 * @throws	ProtocolError
	 * @throws	TransmissionError
	 */
	protected FileList(byte[] datagram) throws TransmissionError, SizeError, ProtocolError, InternalError {
		super(datagram);
		/* assert to help debugging */
		assert requestResponseCode == RequestResponseCode.LIST_RESPONSE : "FileList subclass is incompatible with this datagram, request/response code must be checked before using this constructor";
		/* InternalErrorException */
		if (requestResponseCode!= RequestResponseCode.LIST_RESPONSE) {
			throw new InternalError();
		}
		int size = getPayloadSize(datagram);
		try {
			fileList = (new String(datagram, 8, size, "UTF-8")).split("\n");
		} catch (UnsupportedEncodingException e) {
			throw new InternalError();
		}
	}

	/** Returns a byte[] containing datagram with padding.
	 * This datagram is still incomplete and should not be send directly.
	 * ProtocolP2PDatagram will use this method to generate the complete datagram.
	 * @return datagram with padding
	 * @throws	InternalError
	 */
	protected byte[] toDatagram() throws InternalError {
		// compute size
		int size = 8;
		for (String s : fileList) {
			size += s.length();
			size += 1;
		}
		size -=1;
		byte[] datagram = new byte[size]; // java initialize all to zero
		// set request/response code
		datagram[RequestResponseCode.RRCODE_POSITION] = requestResponseCode.codeValue;
		// bits 16-31 are reserved for future use
		setPayloadSize(size - 8, datagram);
		// Write fileList
		int bCount = 8;
		for(String s : fileList) {
			if (bCount != 8) { // not on first iteration
				try {
					datagram[bCount] = "\n".getBytes("UTF-8")[0]; // separator
				} catch (UnsupportedEncodingException e) {
					throw new InternalError();
				}
				bCount += 1;
			}
			// Copy filename
			try {
				byte[] sb = s.getBytes("UTF-8");
				for(byte b : sb) {
					datagram[bCount] = b;
					bCount += 1;
				}
			} catch (UnsupportedEncodingException e) {
				throw new InternalError();
			}
			
		}
		return datagram;
	}

	/** fileList getter.
	 * @return	fileList
	 */
	public String[] getFileList() {
		return fileList;
	}
}
