package protocolP2P;
import exception.InternalError;
import exception.ProtocolError;
import exception.SizeError;
import exception.TransmissionError;
import exception.VersionError;
import remoteException.EmptyDirectory;
import remoteException.InternalRemoteError;
import remoteException.NotFound;
import remoteException.ProtocolRemoteError;
import remoteException.VersionRemoteError;
import remoteException.EmptyFile;
import tools.BytesArrayTools;
import protocolP2P.Payload;
import protocolP2P.RequestResponseCode;
import protocolP2P.LoadRequest;
import protocolP2P.FileList;
import protocolP2P.FilePart;
import java.util.ArrayList;
import java.lang.Byte;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.io.IOException;
import java.net.UnknownHostException;

/** Representation of datagram.
 * @author	Louis Royer
 * @author	Flavien Haas
 * @author	JS Auge
 * @version	1.0
 */
public class ProtocolP2PDatagram {
	private final static byte PROTOCOL_VERSION = 0x11;
	private final static int VERSION_POSITON = 0;
	private byte version;
	private Payload payload;
	private InetAddress hostR;
	private int portR;
	private final static int CHECKSUM_POSITION = 2;

	/** Constructor with payload parameter (typically used when sending datagram).
	 * @param payload	the payload associated with the datagram to send
	 */
	public ProtocolP2PDatagram(Payload payload) {
		version = PROTOCOL_VERSION;
		this.payload = payload;
	}

	/** Send datagram on socket (from client)
	 * @param socket	DatagramSocket used to send datagram.
	 * @param host	host to send datagram (null if this is a response)
	 * @param port	port to send datagram (null if this is a response)
	 * @throws	InternalError
	 * @throws	UnknownHostException
	 * @throws	IOException
	 */
	public void send(DatagramSocket socket, String host, int port) throws InternalError, UnknownHostException, IOException {
		InetAddress dst = InetAddress.getByName(host);
		byte[] datagram = toDatagram();
		// generate DatagramPacket
		DatagramPacket datagramPacket = new DatagramPacket(datagram, datagram.length, dst, port);
		// send it
		socket.send(datagramPacket);
	}

	/** Send datagram on socket (from server, as a response)
	 * @param socket	DatagramSocket used to send datagram.
	 * @param received	datagram to respond (aka request)
	 * @throws	InternalError
	 * @throws	IOException
	 */
	public void send(DatagramSocket socket, ProtocolP2PDatagram received) throws InternalError, IOException {
		assert received.getPortR() != 0 && received.getHostR() != null : "This method should be used only as response to a request";
		if (received.getPortR() == 0 || received.getHostR() == null) {
			throw new InternalError();
		}
		byte[] datagram = toDatagram();
		// generate DatagramPacket
		DatagramPacket datagramPacket = new DatagramPacket(datagram, datagram.length, received.getHostR(), received.getPortR());
		socket.send(datagramPacket);
	}

	/** Send a response.
	 * @param socket	DatagramSocket used to send response
	 * @param host	host to send response
	 * @param port	port to send response
	 * @throws	InternalError
	 * @throws	IOException
	 */
	protected void sendResponse(DatagramSocket socket, InetAddress host, int port) throws InternalError, IOException {
		assert port != 0 && host != null : "This method should be used only as response to a request";
		if (port == 0 || host == null) {
			throw new InternalError();
		}
		byte[] datagram = toDatagram();
		DatagramPacket datagramPacket = new DatagramPacket(datagram, datagram.length, host, port);
		socket.send(datagramPacket);
	}

	/** Receive datagram on socket
	 * @param socket	DatagramSocket used to receive datagram
	 * @throws	TransmissionError
	 * @throws	ProtocolError
	 * @throws	VersionError
	 * @throws	InternalError
	 * @throws	SizeError
	 * @throws	ProtocolRemoteError
	 * @throws	VersionRemoteError
	 * @throws	InternalRemoteError
	 * @throws	EmptyDirectory
	 * @throws	NotFound
	 * @throws	IOException
	 * @throws	EmptyFile
	 */
	public static ProtocolP2PDatagram receive(DatagramSocket socket) throws EmptyFile, NotFound, EmptyDirectory, InternalRemoteError, VersionRemoteError, ProtocolRemoteError, TransmissionError, ProtocolError, VersionError, InternalError, SizeError, IOException {
		// reception
		byte[] datagram = new byte[4096];
		DatagramPacket reception = new DatagramPacket(datagram, datagram.length);
		socket.receive(reception);
		// contruction
		try {
			ProtocolP2PDatagram p = new ProtocolP2PDatagram(datagram);
			p.setHostR(reception.getAddress());
			p.setPortR(reception.getPort());
			Payload payload = p.getPayload();
			switch (payload.getRequestResponseCode()) {
				case PROTOCOL_ERROR :
					throw new ProtocolRemoteError();
				case VERSION_ERROR :
					throw new VersionRemoteError();
				case INTERNAL_ERROR :
					throw new InternalRemoteError();
				case EMPTY_DIRECTORY :
					throw new EmptyDirectory();
				case NOT_FOUND :
					throw new NotFound();
				case EMPTY_FILE:
					throw new EmptyFile();
				default :
					return p;
			}
		} catch (TransmissionError e) {
			(new ProtocolP2PDatagram(new Payload(RequestResponseCode.INTERNAL_ERROR))).sendResponse(socket, reception.getAddress(), reception.getPort());
			throw e;
		} catch (ProtocolError e) {
			(new ProtocolP2PDatagram(new Payload(RequestResponseCode.PROTOCOL_ERROR))).sendResponse(socket, reception.getAddress(), reception.getPort());
			throw e;
		} catch (VersionError e) {
			(new ProtocolP2PDatagram(new Payload(RequestResponseCode.VERSION_ERROR))).sendResponse(socket, reception.getAddress(), reception.getPort());
			throw e;
		} catch (InternalError e) {
			(new ProtocolP2PDatagram(new Payload(RequestResponseCode.INTERNAL_ERROR))).sendResponse(socket, reception.getAddress(), reception.getPort());
			throw e;
		} catch (SizeError e) {
			(new ProtocolP2PDatagram(new Payload(RequestResponseCode.INTERNAL_ERROR))).sendResponse(socket, reception.getAddress(), reception.getPort());
			throw e;
		}
	}
	/** Private constructor with datagram as byte[] parameter (typically used when receiving datagram).
	 * @param datagram	the full datagram received
	 * @throws	TransmissionError
	 * @throws	ProtocolError
	 * @throws	VersionError
	 * @throws	InternalError
	 * @throws	SizeError
	 */
	private ProtocolP2PDatagram(byte[] datagram) throws TransmissionError, ProtocolError, VersionError, InternalError, SizeError {
		// unwrap version
		version = datagram[VERSION_POSITON];
		checkProtocolVersion(); // this can throw VersionError
		checkCheckSum(datagram);
		RequestResponseCode r = RequestResponseCode.fromCode(datagram[RequestResponseCode.RRCODE_POSITION]); // this can throw ProtocolError
		switch (r) {
			case LIST_RESPONSE:
				payload = (Payload) new FileList(datagram);
				break;
			case LOAD_RESPONSE:
				payload = (Payload) new FilePart(datagram);
				break;
			case LOAD_REQUEST:
				payload = (Payload) new LoadRequest(datagram);
				break;
			default:
				payload = new Payload(datagram); // this can throw TransmissionError
				break;
			}
	}

	/** Returns a byte[] containing full datagram (typically used when sending datagram).
	 * This datagram is still complete and ready to be send.
	 * @return the full datagram to send
	 * @throws	InternalError
	 */
	protected byte[] toDatagram() throws InternalError {
		byte[] datagram = payload.toDatagram();
		datagram[VERSION_POSITON] = version;
		setCheckSum(datagram);
		return datagram;
	}

	/** Returns Payload associated with the datagram.
	 * @return	payload associated with the datagram
	 */
	public Payload getPayload() {
		return payload;
	}

	/** Used to check protocol version when a datagram is constructed from bytes[].
	 * @throws	VersionError
	 */
	private void checkProtocolVersion() throws VersionError {
		if (PROTOCOL_VERSION != version) {
			throw new VersionError();
		}
	}

	/** portR getter.
	 * @return	portR
	 */
	protected int getPortR() {
		return portR;
	}

	/** portR setter.
	 * @param portR	portR
	 */
	 protected void setPortR(int portR) {
		 this.portR = portR;
	 }

	/** hostR getter.
	 * @return	hostR
	 */
	protected InetAddress getHostR() {
		return hostR;
	}

	/** hostH setter.
	 * @param hostR	hostR
	 */
	protected void setHostR(InetAddress hostR) {
		this.hostR = hostR;
	}

	private int computeCheckSum(byte [] datagram){
		int checksum = 0;
		for (int i=CHECKSUM_POSITION+2; i<datagram.length; ++i){
			checksum += BytesArrayTools.readInt16Bits(datagram,i);
			checksum &= 0xffff;
		}
		return checksum ^ 0xffff;
	}

	private void setCheckSum(byte [] datagram) throws InternalError{
		int checksum = computeCheckSum(datagram);
		try{
			BytesArrayTools.write16Bits(datagram,CHECKSUM_POSITION,checksum);
		} catch (SizeError e) {
			throw new InternalError();
		}
	}

	private void checkCheckSum(byte [] datagram) throws TransmissionError{
		int checksum = BytesArrayTools.readInt16Bits(datagram,CHECKSUM_POSITION);
		int computedCheckSum = computeCheckSum(datagram);
		if (computedCheckSum != checksum){
			throw new TransmissionError();
		}
	}
}
