package protocolP2P;
import protocolP2P.RequestResponseCode;
import protocolP2P.FilePart;
import protocolP2P.FileList;
import protocolP2P.LoadRequest;
import exception.ProtocolError;
import exception.InternalError;
import exception.TransmissionError;
import exception.SizeError;
import tools.BytesArrayTools;
/** Representation of payload. If payload has a size, use subclasses instead.
 * @author	Louis Royer
 * @author	Flavien Haas
 * @author	JS Auge
 * @version	1.0
 */
public class Payload {
	protected RequestResponseCode requestResponseCode;
	protected final static int PAYLOAD_SIZE_POSITION = 4;
	protected final static int PAYLOAD_START_POSITION = 8;

	/** Consructor used to create Payload with a payload size of zero using a RRCode.
	 * @param requestResponseCode	Request/Response code associated with the payload
	 * @throws	InternalError
	 */
	public Payload(RequestResponseCode requestResponseCode) throws InternalError {
		/* asserts to help debugging */
		assert requestResponseCode != RequestResponseCode.LIST_RESPONSE || (this instanceof FileList) : "LIST_RESPONSE must use FilePart class";
		assert requestResponseCode != RequestResponseCode.LOAD_RESPONSE || (this instanceof FilePart) : "LOAD_RESPONSE must use FileList class";
		assert requestResponseCode != RequestResponseCode.LOAD_REQUEST  || (this instanceof LoadRequest) : "LOAD_REQUEST must use LoadRequest class";
		this.requestResponseCode = requestResponseCode;
		checkRequestResponseCode(); // this can throw InternalError
	}

	/** Constructor used to create a Payload (when no more specific subclasses exists) using datagram as parameter.
	 * If payload size is not empty, using subclass is required.
	 * @param datagram	the full datagram received
	 * @throws	ProtocolError
	 * @throws	InternalError
	 * @throws	TransmissionError
	 * @throws	SizeError
	 */
	protected Payload(byte[] datagram) throws SizeError, ProtocolError, InternalError, TransmissionError {
		/* asserts to help debugging */
		assert getPayloadSize(datagram) + 8 <= datagram.length : "Payload is truncated";
		if (datagram.length < getPayloadSize(datagram) + 8) {
			throw new TransmissionError();
		}
		assert RequestResponseCode.fromCode(datagram[RequestResponseCode.RRCODE_POSITION]) != RequestResponseCode.LIST_RESPONSE || (this instanceof FileList) : "LIST_RESPONSE must use FilePart class";
		assert RequestResponseCode.fromCode(datagram[RequestResponseCode.RRCODE_POSITION]) != RequestResponseCode.LOAD_RESPONSE || (this instanceof FilePart) : "LOAD_RESPONSE must use FileList class";
		assert RequestResponseCode.fromCode(datagram[RequestResponseCode.RRCODE_POSITION]) != RequestResponseCode.LOAD_REQUEST  || (this instanceof LoadRequest) : "LOAD_REQUEST must use LoadRequest class";
		requestResponseCode = RequestResponseCode.fromCode(datagram[RequestResponseCode.RRCODE_POSITION]);
		checkRequestResponseCode(); // this can throw InternalError
	}

	/** Used to check RRCode used is compatible with this class use, or if a more specific subclass is required.
	 * @throws	InternalError
	 */ 
	private void checkRequestResponseCode() throws InternalError {
		/* Incorrect use cases (use subclasses instead) */
		if ((requestResponseCode == RequestResponseCode.LIST_RESPONSE && !(this instanceof FileList))
		 || (requestResponseCode == RequestResponseCode.LOAD_RESPONSE && !(this instanceof FilePart))
		 || (requestResponseCode == RequestResponseCode.LOAD_REQUEST  && !(this instanceof LoadRequest))) {
			throw new InternalError();
		}
	}
	
	/** Returns a byte[] containing datagram with padding.
	 * This datagram is still incomplete and should not be send directly.
	 * ProtocolP2PDatagram will use this method to generate the complete datagram.
	 * @return datagram with padding
	 * @throws	InternalError
	 */
	protected byte[] toDatagram() throws InternalError {
		// InternalError is impossible in this method on Payload class but still on subclasses
		byte [] datagram = new byte[8]; // java initialize all to zero
		// size is zero (and this is the default)
		// set request/response code
		datagram[RequestResponseCode.RRCODE_POSITION] = requestResponseCode.codeValue;
		// bits 16-31 are reserved for future use
		// payload size is 0 (this is what java have initialized datagram)
		return datagram;
	}

	/** Set payload’s size in a datagram.
	 * @param size	integer representing payload size
	 * @param datagram	datagram to be completed
	 * @throws	InternalError
	 */ 
	protected static void setPayloadSize(int size, byte[] datagram) throws InternalError {
		/* assert to help debugging */
		assert size >= 0: "Payload size cannot be negative";
		if (size < 0) {
			// We don't throw SizeError
			// because this is only for reception side
			throw new InternalError();
		}
		BytesArrayTools.write(datagram, PAYLOAD_SIZE_POSITION, size);
	}

	/** Get payload’s size from a datagram.
	 * @param datagram	the full datagram received
	 * @return	integer representing payload size
	 * @throws	SizeError
	 */
	protected static int getPayloadSize(byte[] datagram) throws SizeError {
		return BytesArrayTools.readInt(datagram, PAYLOAD_SIZE_POSITION);
	}

	/** RRCode getter.
	 * @return	Request/Response code
	 */
	public RequestResponseCode getRequestResponseCode() {
		return requestResponseCode;
	}
}
